(*
    Differencing engine
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit DiffEngine;

interface

uses
  Classes;

type
  TDiffEngine = class
    FIgnoreBlanks: boolean;
    FIgnoreCase: boolean;

    function LinesAreEqual( const L1, L2: string ): boolean;
    function GetStringCode( S: string ): cardinal;
    function PrepareString( S: string ): string;

  public
    constructor Create;

    function Diff( const ADoc1, ADoc2: TStringList; ADiffs: TList ): integer;

    // Merges a document with a diff list
    function Merge( ADoc: TStringList; ADiffs: TList; AReverse: boolean ): integer;

    class function FreeDiffList( ADiffs: TList ): integer;
    class function MakeSimpleDiffList( const ADoc1, ADoc2: TStringList; ADiffs: TList ): integer;

    property IgnoreBlanks: boolean read FIgnoreBlanks write FIgnoreBlanks;
    property IgnoreCase: boolean read FIgnoreCase write FIgnoreCase;
  end;

implementation

uses
  SysUtils, Math,
  DiffTypes;

const
  BLANKS = [ #9, ' ' ];

function TDiffEngine.LinesAreEqual( const L1, L2: string ): boolean;
begin
  Result := L1 = L2;
end;

constructor TDiffEngine.Create;
begin
  FIgnoreBlanks := false;
  FIgnoreCase := false;
end;

function TDiffEngine.PrepareString( S: string ): string;
begin
  if( FIgnoreBlanks ) then
    S := Trim( S );

  if( FIgnoreCase ) then
    S := AnsiLowerCase( S );

  Result := S;
end;

function TDiffEngine.GetStringCode( S: string ): cardinal;
var
  L: integer;
begin
  L := Length(S);

  Result := (L and $7F) shl 25;

  if( Length(S) >= 1 ) then Result := Result or ((Ord(S[1]) and $1F) shl 20);
  if( Length(S) >= 2 ) then Result := Result or ((Ord(S[2]) and $1F) shl 15);
  if( Length(S) >= 3 ) then Result := Result or ((Ord(S[L div 2]) and $1F) shl 10);
  if( Length(S) >= 4 ) then Result := Result or ((Ord(S[L div 3]) and $1F) shl  5);
  if( Length(S) >= 5 ) then Result := Result or ((Ord(S[L]) and $1F) shl  0);
end;

(*
    This function computes the differences between the two
    specified files and returns them as a list of TDiffLine items.

    Note: this is a very simple implementation that is definitely
    *not* designed for robustness and speed. However, it is very
    easy to read and understand, which is the main goal.
*)
function TDiffEngine.Diff( const ADoc1, ADoc2: TStringList; ADiffs: TList ): integer;
var
  L1, L2: integer;
  PL: array of array of smallint;
  I, J: integer;
  LineNo: integer;
  D1, D2: array of cardinal;
  Z1, Z2: TStringList;
procedure AddDiff( M: TDiffLineMode; L1, L2: integer );
  begin
    ADiffs.Add( TDiffLine.Create( M, L1, L2, LineNo ) );
    Inc( LineNo );
  end;
begin
  Result := 0;

  L1 := 0;
  L2 := 0;
  if( ADoc1 <> nil ) then L1 := ADoc1.Count;
  if( ADoc2 <> nil ) then L2 := ADoc2.Count;

  if( L1 = 0 )and( L2 = 0 ) then begin
    // Nothing to do!
    Exit;
  end;

  // Allocate the table for the partial lengths
  SetLength( PL, L1+1, L2+1 );

  // Apply basic transformation to the input documents so that we
  // don't have to inside the loops
  Z1 := TStringList.Create;
  Z2 := TStringList.Create;
  for I:=1 to L1 do
    Z1.Add( PrepareString(ADoc1[I-1]) );
  for I:=1 to L2 do
    Z2.Add( PrepareString(ADoc2[I-1]) );

  // Compute a simple code for each string that represents
  // the string length and the first few characters of the string:
  // this code will *enormously* speed up the comparisons later
  SetLength( D1, L1 );
  SetLength( D2, L2 );
  for I:=1 to L1 do
    D1[I-1] := GetStringCode( Z1[I-1] );
  for I:=1 to L2 do
    D2[I-1] := GetStringCode( Z2[I-1] );

  // Initialize the extra column and row to zero
  for I:=0 to L1 do
    PL[I,L2] := 0;

  for J:=0 to L2 do
    PL[L1,J] := 0;

  // Compute the Longest Common Subsequence (LCS) starting from the end
  // so we can reuse partial result as we move towards the beginning
  for I:=L1-1 downto 0 do begin
    for J:=L2-1 downto 0 do begin
       // Note: this optimization relies on the fact that if the first
       // condition is false then the compiler does not evaluate the
       // second condition, which is *much* slower
       if (D1[I] = D2[J]) and ((D1[I] = 0)or(Z1[I] = Z2[J])) then begin
        // Since the two strings match, the line is part of
        // the LCS, i.e. we add one to the LCS length
        PL[I,J] := 1 + PL[I+1,J+1];
      end
      else begin
        // Since the two strings do not match, we have to choose
        // which line has to be included in the LCS: we do this
        // by trying both cases and then choosing the one that leads
        // to the longest subsequence
        PL[I,J] := Max( PL[I+1,J], PL[I,J+1] );
      end;
    end;
  end;

  // Use the partial length table to build the difference sequence
  I := 0;
  J := 0;
  LineNo := 0;

  while( I < L1 )and( J < L2 ) do begin
    if (D1[I] = D2[J]) and ((D1[I] = 0)or(Z1[I] = Z2[J])) then begin
      // This line is the same in both files, add it to the result
      // and increment both line counters
      AddDiff( dmBothDocuments, I, J );
      Inc( I );
      Inc( J );
    end
    else if( PL[I+1,J] > PL[I,J+1] ) then begin
      // This line is different and choosing file one yields
      // a longer subsequence, so add it to the first file and
      // increment the line counter accordingly
      AddDiff( dmOnlyDoc1, I, -1 );
      Inc( I );
    end
    else begin
      AddDiff( dmOnlyDoc2, -1, J );
      Inc( J );
    end;
  end;

  // Add trailing lines that might be excluded by the previous loop
  while( I < L1 ) do begin
    AddDiff( dmOnlyDoc1, I, -1 );
    Inc( I );
  end;
  while( J < L2 ) do begin
    AddDiff( dmOnlyDoc2, -1, J );
    Inc( J );
  end;
end;

function TDiffEngine.Merge( ADoc: TStringList; ADiffs: TList; AReverse: boolean ): integer;
var
  I: integer;
  D: TSimpleDiffLine;
  L: integer;
begin
  Result := 0;

  if( AReverse ) then begin
    for I:=ADiffs.Count downto 1 do begin
      D := TSimpleDiffLine( ADiffs[I-1] );
      L := D.LineNo;
      if( D.Mode = dmOnlyDoc2 ) then begin
        // Line must be inserted into document
        ADoc.Insert( L, D.Line );
        Inc( Result );
      end
      else begin
        // Line must be deleted from document
        ADoc.Delete( L );
        Inc( Result );
      end;
    end;
  end
  else begin
    for I:=1 to ADiffs.Count do begin
      D := TSimpleDiffLine( ADiffs[I-1] );
      L := D.LineNo;
      if( D.Mode = dmOnlyDoc1 ) then begin
        // Line must be inserted into document
        ADoc.Insert( L, D.Line );
        Inc( Result );
      end
      else begin
        // Line must be deleted from document
        ADoc.Delete( L );
        Inc( Result );
      end;
    end;
  end;
end;

class function TDiffEngine.MakeSimpleDiffList( const ADoc1, ADoc2: TStringList; ADiffs: TList ): integer;
var
  I: integer;
  D: TDiffLine;
  L: TList;
  N: integer;
begin
  Result := 0;

  N := 0;
  L := TList.Create;
  for I:=1 to ADiffs.Count do begin
    D := TDiffLine( ADiffs[I-1] );
    case D.Mode of
      dmOnlyDoc1: L.Add( TSimpleDiffLine.Create( D.Mode, ADoc1[D.LineNo1], D.LineNo1 ) );
      dmOnlyDoc2: L.Add( TSimpleDiffLine.Create( D.Mode, ADoc2[D.LineNo2], N ) );
    end;
    if( D.Mode in [dmOnlyDoc1,dmBothDocuments] ) then
      Inc( N );
    D.Free;
  end;
  ADiffs.Clear;
  for I:=1 to L.Count do
    ADiffs.Add( L[I-1] );
  L.Free;
end;

class function TDiffEngine.FreeDiffList( ADiffs: TList ): integer;
var
  I: integer;
  D: TObject;
begin
  Result := 0;

  if( ADiffs <> nil ) then begin
    Result := ADiffs.Count;
    for I:=ADiffs.Count downto 1 do begin
      D := TObject( ADiffs[I-1] );
      D.Free;
    end;
    ADiffs.Free;
  end;
end;

end.
