(*
    Differencing engine
    Copyright (c) 2002-2003 Alessandro Scotti
    http://www.walkofmind.com/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit DiffFileEngine;

interface

uses Classes, DiffLine;

type
  TDiffOptions = record
    IgnoreBlanks: boolean;
  end;

function diffStringLists( var T1, T2: TStringList; IgnoreBlanks: boolean ): TDiffLineList;
function diffTextFiles( const FilenameOne, FilenameTwo: string; const Options: TDiffOptions ): TDiffLineList;

implementation

uses SysUtils, Dialogs, Math;

function LoadFile( FileName: string; SL: TStringList ): boolean;
begin
  Result := false;
  if( FileName <> '' ) then begin
    try
      SL.LoadFromFile( FileName );
      Result := true;
    except
      on E: Exception do
        MessageDlg( 'Cannot load ''' + FileName + '''. ' + E.Message + '.', mtError, [mbOk], 0 );
    else
        MessageDlg( 'Cannot load ''' + FileName + '''. ', mtError, [mbOk], 0 );
    end;
  end;
end;

function diffTextFiles( const FilenameOne, FilenameTwo: string; const Options: TDiffOptions ): TDiffLineList;
var
  T1, T2: TStringList;
begin
  T1 := nil;
  T2 := nil;
  try
    // Load the input files in memory
    T1 := TStringList.Create;
    T2 := TStringList.Create;
    LoadFile( FileNameOne, T1 );
    LoadFile( FileNameTwo, T2 );
    // Get the difference
    Result := diffStringLists( T1, T2, Options.IgnoreBlanks );
  except
    Result := nil;
  end;
  T1.Free;
  T2.Free;
end;

const
  BLANKS = [ #9, ' ' ];

function LinesAreEqual( const L1, L2: string; IgnoreBlanks: boolean ): boolean;
var
  P1, P2: PChar;
begin
  if( not IgnoreBlanks ) then begin
    Result := L1 = L2;
  end
  else begin
    P1 := PChar(L1);
    P2 := PChar(L2);
    while( P1^ <> #0 )and( P2^ <> #0 ) do begin
      if( P1^ in BLANKS ) then begin
        if not( P2^ in BLANKS ) then
          break;
        repeat
          Inc( P1 );
        until not (P1^ in BLANKS);
        repeat
          Inc( P2 );
        until not (P2^ in BLANKS); 
      end
      else begin
        if( P1^ <> P2^ ) then
          break;
        Inc( P1 );
        Inc( P2 );
      end;
    end;

    // Ignore trailing blanks
    if( P1^ = #0 ) then while( P2^ in BLANKS ) do Inc( P2 );
    if( P2^ = #0 ) then while( P1^ in BLANKS ) do Inc( P1 );

    Result := P1^ = P2^;
  end;
end;

(*
    This function computes the differences between the two
    specified files and returns them in a single document
    of type TDiffLineList.

    Note: this is a very simple implementation that is definitely
    *not* designed for robustness and speed. However, it is very
    easy to read and understand, which is the main goal.
*)
function diffStringLists( var T1, T2: TStringList; IgnoreBlanks: boolean ): TDiffLineList;
var
  L1, L2: integer;
  PL: array of array of smallint;
  I, J: integer;
begin
  L1 := T1.Count;
  L2 := T2.Count;

  Result := nil;
  if( L1 = 0 )and( L2 = 0 ) then // Nothing to do!
    Exit;

  // Allocate the table for the partial lengths
  SetLength( PL, L1+1, L2+1 );

  // Compute the Longest Common Subsequence (LCS) starting from the end
  // so we can reuse partial result as we move towards the beginning
  for I:=0 to L1 do
    PL[I,L2] := 0;
  for J:=0 to L2 do
    PL[L1,J] := 0;
  for I:=L1-1 downto 0 do begin
    for J:=L2-1 downto 0 do begin
       if LinesAreEqual( T1[I], T2[J], IgnoreBlanks ) then begin
        // Since the two strings match, the line is part of
        // the LCS, i.e. we add one to the LCS length
        PL[I,J] := 1 + PL[I+1,J+1];
      end
      else begin
        // Since the two strings do not match, we have to choose
        // which line has to be included in the LCS: we do this
        // by trying both cases and then choosing the one that leads
        // to the longest subsequence
        PL[I,J] := Max( PL[I+1,J], PL[I,J+1] );
      end;
    end;
  end;

  // Use the partial length table to build the difference sequence
  Result := TDiffLineList.Create;

  I := 0;
  J := 0;

  while( I < L1 )and( J < L2 ) do begin
    if LinesAreEqual( T1[I], T2[J], IgnoreBlanks ) then begin
      // This line is the same in both files, add it to the result
      // and increment both line counters
      Result.Add( dmBothFiles, T2[J] );
      Inc( I );
      Inc( J );
    end
    else if( PL[I+1,J] >= PL[I,J+1] ) then begin
      // This line is different and choosing file one yields
      // a longer subsequence, so add it to the first file and
      // increment the line counter accordingly
      Result.Add( dmFileOneOnly, T1[I] );
      Inc( I );
    end
    else begin
      Result.Add( dmFileTwoOnly, T2[J] );
      Inc( J );
    end;
  end;

  // Add trailing lines that might be excluded by the previous loop
  while( I < L1 ) do begin
    Result.Add( dmFileOneOnly, T1[I] );
    Inc( I );
  end;
  while( J < L2 ) do begin
    Result.Add( dmFileTwoOnly, T2[J] );
    Inc( J );
  end;
end;

end.
