(*
    Internal representation of differenced files
    Copyright (c) 2002-2003 Alessandro Scotti
    http://www.walkofmind.com/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit DiffLine;

interface

uses Classes;

type
  TDiffLineMode = (
    dmNone,                 // Unassigned
    dmFileOneOnly,          // Line is in first file only
    dmFileTwoOnly,          // Line is in second file only
    dmBothFiles,            // Line is in both files
    // Can only get the following after some user action
    dmCopiedFromFileOne,    // Line was copied from first file to second file (so it's now in both files)
    dmCopiedFromFileTwo,    // Line was copied from second file to first file (so it's now in both files)
    dmDeletedFromFileOne,   // Line was deleted from first file (by the user, it's kept in the document for the undo)
    dmDeletedFromFileTwo    // Line was deleted from second file (by the user, it's kept in the document for the undo)
  );

type
  (*
    This is a simple class that represent one line in the
    difference document.
    It stores one line of text plus the information about
    which of the input file generated this line. Upon creation
    of a new difference document the line mode can only assume
    three values: dmBothFiles, dmFileOneOnly and dmFileTwoOnly.
    Other values may be assumed later when editing the document.
  *)
  TDiffLine = class
  private
    FMode: TDiffLineMode;   // Line mode (see above)
    FLine: string;          // Line text
  public
    constructor Create;
    constructor CreateFrom( AMode: TDiffLineMode; ALine: string );

    property Mode: TDiffLineMode read FMode write FMode;
    property Line: string read FLine write FLine;
  end;

  (*
    This represents the output of differencing two files.
    It is simply a list of TDiffLine's (see above) and hold enough
    information to recreate both of the input files from this
    single source.
  *)
  TDiffLineList = class
  private
    FList: TList;

    function GetCount: integer;
    function GetLine( AIndex: integer ): TDiffLine;
  public
    constructor Create;
    destructor Destroy; override;

    function Add( AMode: TDiffLineMode; ALine: string ): integer; // Adds a line to the end of the list

    property Count: integer read GetCount; // Number of lines in the list
    property Lines[AIndex: integer]: TDiffLine read GetLine; // Get line at specified index
  end;

implementation

// Creates an empty line
constructor TDiffLine.Create;
begin
  FMode := dmNone;
  FLine := '';
end;

// Creates a line from specified values
constructor TDiffLine.CreateFrom( AMode: TDiffLineMode; ALine: string );
begin
  FMode := AMode;
  FLine := ALine;
end;

constructor TDiffLineList.Create;
begin
  FList := TList.Create;
end;

destructor TDiffLineList.Destroy;
var
  I: integer;
  P: TDiffLine;
begin
  if( FList <> nil ) then begin
    for I:=0 to FList.Count-1 do begin
      P := TDiffLine( FList.Items[I] );
      P.Free;
    end;
    FList.Free;
  end;
  inherited Destroy;
end;

function TDiffLineList.GetCount: integer;
begin
  Result := FList.Count;
end;

function TDiffLineList.GetLine( AIndex: integer ): TDiffLine;
var
  P: pointer;
begin
  P := nil;
  if( AIndex >= 0 )and( AIndex < FList.Count ) then begin
    P := FList.Items[ AIndex ];
  end;

  Result := TDiffLine( P );
end;

function TDiffLineList.Add( AMode: TDiffLineMode; ALine: string ): integer;
var
  P: TDiffLine;
begin
  P := TDiffLine.CreateFrom( AMode, ALine );
  Result := FList.Add( P );
end;

end.
