(*
    Basic types for differencing text documents
    Copyright (c) 2003 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit DiffTypes;

interface

uses
  Classes, DiffEngine;

type
  TDiffLineMode = (
    dmNone,                 // Unassigned
    dmOnlyDoc1,             // Line is in first file only
    dmOnlyDoc2,             // Line is in second file only
    dmBothDocuments,        // Line is in both files
    // Can only get the following after some user action
    dmCopiedFromDoc1,       // Line was copied from first file to second file (so it's now in both files)
    dmCopiedFromDoc2,       // Line was copied from second file to first file (so it's now in both files)
    dmDeletedFromDoc1,      // Line was deleted from first file (by the user, it's kept in the document for the undo)
    dmDeletedFromDoc2       // Line was deleted from second file (by the user, it's kept in the document for the undo)
  );

type
  TDiffDocument = class;

  TSimpleDiffLine = class
    FMode: TDiffLineMode;
    FLine: string;
    FLineNo: integer;
    FData: TObject;

  public
    constructor Create( AMode: TDiffLineMode; ALine: string; ALineNo: integer );

    property Mode: TDiffLineMode read FMode write FMode;
    property LineNo: integer read FLineNo;
    property Line: string read FLine;
    property Data: TObject read FData write FData;
  end;

  TDiffLine = class
    FMode: TDiffLineMode;   // Type of difference
    FLineNo1: integer;      // Index of text line in original document 1
    FLineNo2: integer;      // Index of text line in original document 2
    FOwner: TDiffDocument;
    FRawLineNo: integer;    // Index of difference in raw (unedited) difference list
    FData: TObject;

    function GetLine1: string;
    function GetLine2: string;

  public
    constructor Create( AMode: TDiffLineMode; ALineNo1, ALineNo2, ARawLineNo: integer );

    property Mode: TDiffLineMode read FMode write FMode;
    property LineNo1: integer read FLineNo1;
    property LineNo2: integer read FLineNo2;
    property Line1: string read GetLine1;
    property Line2: string read GetLine2;
    property Owner: TDiffDocument read FOwner write FOwner;
    property RawLineNo: integer read FRawLineNo;
    property Data: TObject read FData write FData;
  end;

  TDiffDocument = class
    FDoc1: TStringList;         // Original document 1
    FDoc2: TStringList;         // Original document 2
    FRawDiffs: TList;           // Raw (unedited) list of differences
    FDiffs: TList;              // Current list of differences
    FCurDoc1: TList;            // Current document 1 (original plus edited differences)
    FCurDoc2: TList;            // Current document 2 (original plus edited differences)
    FDiffEngine: TDiffEngine;
    FDiffCount1: integer;
    FDiffCount2: integer;

    function GetCount: integer;
    function GetLine( AIndex: integer ): TDiffLine;
    function GetMode( AIndex: integer ): TDiffLineMode;
    function GetRawCount: integer;
    function GetRawLine( AIndex: integer ): TDiffLine;
    function GetDocLineCount1: integer;
    function GetDocLineCount2: integer;
    function GetDocLine1( AIndex: integer ): TDiffLine;
    function GetDocLine2( AIndex: integer ): TDiffLine;

    procedure Clear;

  public
    constructor Create;
    destructor Destroy; override;

    // Rebuilds the difference document
    procedure RebuildAll;

    // Rebuilds the current document
    procedure RebuildDiff;

    // Source documents
    property RawDoc1: TStringList read FDoc1 write FDoc1;
    property RawDoc2: TStringList read FDoc2 write FDoc2;

    // Full document, includes deleted lines
    property RawLine[AIndex: integer]: TDiffLine read GetRawLine;
    property RawCount: integer read GetRawCount;

    // Current document
    property Line[AIndex: integer]: TDiffLine read GetLine;
    property Mode[AIndex: integer]: TDiffLineMode read GetMode;
    property Count: integer read GetCount;
    property DiffCount1: integer read FDiffCount1;
    property DiffCount2: integer read FDiffCount2;
    property DiffLineCount1: integer read GetDocLineCount1;
    property DiffLineCount2: integer read GetDocLineCount2;
    property DiffLine1[AIndex: integer]: TDiffLine read GetDocLine1;
    property DiffLine2[AIndex: integer]: TDiffLine read GetDocLine2;

    // Differencing engine
    property DiffEngine: TDiffEngine read FDiffEngine write FDiffEngine;
  end;

function IsDifference( Mode: TDiffLineMode ): boolean;
  
implementation

constructor TDiffLine.Create( AMode: TDiffLineMode; ALineNo1, ALineNo2, ARawLineNo: integer );
begin
  FMode := AMode;
  FLineNo1 := ALineNo1;
  FLineNo2 := ALineNo2;
  FRawLineNo := ARawLineNo;
  FOwner := nil;
end;

function TDiffLine.GetLine1: string;
begin
  Result := '';
  case FMode of
    dmOnlyDoc1, dmBothDocuments, dmCopiedFromDoc1:
      Result := FOwner.RawDoc1[ FLineNo1 ];
    dmCopiedFromDoc2:
      Result := FOwner.RawDoc2[ FLineNo2 ];
  end;
end;

function TDiffLine.GetLine2: string;
begin
  Result := '';
  case FMode of
    dmCopiedFromDoc1:
      Result := FOwner.RawDoc1[ FLineNo1 ];
    dmOnlyDoc2, dmBothDocuments, dmCopiedFromDoc2:
      Result := FOwner.RawDoc2[ FLineNo2 ];
  end;
end;

constructor TDiffDocument.Create;
begin
  inherited Create;
  FRawDiffs := TList.Create;
  FDiffs := TList.Create;
  FCurDoc1 := TList.Create;
  FCurDoc2 := TList.Create;
  FDoc1 := nil;
  FDoc2 := nil;
  FDiffEngine := nil;
  FDiffCount1 := 0;
  FDiffCount2 := 0;
end;

destructor TDiffDocument.Destroy;
begin
  Clear;
  FRawDiffs.Free;
  FDiffs.Free;
  FCurDoc1.Free;
  FCurDoc2.Free;
  inherited Destroy;
end;

function TDiffDocument.GetCount: integer;
begin
  Result := FDiffs.Count;
end;

function TDiffDocument.GetLine( AIndex: integer ): TDiffLine;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FDiffs.Count ) then begin
    Result := TDiffLine( FDiffs[AIndex] );
  end;
end;

function TDiffDocument.GetMode( AIndex: integer ): TDiffLineMode;
var
  D: TDiffLine;
begin
  Result := dmNone;
  D := GetLine(AIndex);
  if( D <> nil ) then begin
    Result := D.Mode;
  end;
end;

function TDiffDocument.GetRawCount: integer;
begin
  Result := FRawDiffs.Count;
end;

function TDiffDocument.GetRawLine( AIndex: integer ): TDiffLine;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FRawDiffs.Count ) then begin
    Result := TDiffLine( FRawDiffs[AIndex] );
  end;
end;

function TDiffDocument.GetDocLineCount1: integer;
begin
  Result := FCurDoc1.Count;
end;

function TDiffDocument.GetDocLineCount2: integer;
begin
  Result := FCurDoc2.Count;
end;

function TDiffDocument.GetDocLine1( AIndex: integer ): TDiffLine;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FCurDoc1.Count ) then
    Result := TDiffLine( FCurDoc1[AIndex] );
end;

function TDiffDocument.GetDocLine2( AIndex: integer ): TDiffLine;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FCurDoc2.Count ) then
    Result := TDiffLine( FCurDoc2[AIndex] );
end;

procedure TDiffDocument.RebuildDiff;
var
  I: integer;
  D: TDiffLine;
begin
  FDiffs.Clear;
  FCurDoc1.Clear;
  FCurDoc2.Clear;
  FDiffCount1 := 0;
  FDiffCount2 := 0;
  for I:=1 to RawCount do begin
    D := RawLine[I-1];
    if not( D.Mode in [dmNone,dmDeletedFromDoc1,dmDeletedFromDoc2] ) then begin
      FDiffs.Add( D );
      if( D.Mode in [dmOnlyDoc1,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then
        FCurDoc1.Add( D );
      if( D.Mode in [dmOnlyDoc2,dmBothDocuments,dmCopiedFromDoc1,dmCopiedFromDoc2] ) then
        FCurDoc2.Add( D );
      case D.Mode of
        dmOnlyDoc1: Inc( FDiffCount1 );
        dmOnlyDoc2: Inc( FDiffCount2 );
      end;
    end;
  end;
end;

procedure TDiffDocument.Clear;
var
  I: integer;
begin
  for I:=1 to RawCount do begin
    RawLine[I-1].Free;
  end;
  FRawDiffs.Clear;
  FDiffs.Clear;
  FCurDoc1.Clear;
  FCurDoc2.Clear;
  FDiffCount1 := 0;
  FDiffCount2 := 0;
end;

procedure TDiffDocument.RebuildAll;
var
  I: integer;
begin
  Clear;
  if( FDiffEngine <> nil ) then begin
    FDiffEngine.Diff( FDoc1, FDoc2, FRawDiffs );
    for I:=1 to RawCount do begin
      TDiffLine(RawLine[I-1]).Owner := Self;
    end;
    RebuildDiff;
  end;
end;

function IsDifference( Mode: TDiffLineMode ): boolean;
begin
  Result := Mode in [dmOnlyDoc1,dmOnlyDoc2];
end;

constructor TSimpleDiffLine.Create( AMode: TDiffLineMode; ALine: string; ALineNo: integer );
begin
  FMode := AMode;
  FLine := ALine;
  FLineNo := ALineNo;
  FData := nil;
end;

end.
