(*
    LCS - A study for a simple differencing program
    Full text editor components
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Editor;

interface

uses
  Classes;

type
  TTextPosition = record
    Col: integer;
    Row: integer;
  end;

  TTextBlockMode = ( bmColumn, bmLinear );

  TTextBlock = class
    FMode: TTextBlockMode;
    FStart: TTextPosition;
    FEnd: TTextPosition;
    FMarkerStart: TTextPosition;
    FMarkerEnd: TTextPosition;
    FData: pointer;

    procedure SetStartMarker( APos: TTextPosition );
    procedure SetEndMarker( APos: TTextPosition );

  public
    constructor Create;

    procedure Clear;

    function FirstRow: integer;
    function LastRow: integer;
    function Valid: boolean;

    property Mode: TTextBlockMode read FMode write FMode;
    property TopLeft: TTextPosition read FStart;
    property BottomRight: TTextPosition read FEnd;
    property StartMarker: TTextPosition read FMarkerStart write SetStartMarker;
    property EndMarker: TTextPosition read FMarkerEnd write SetEndMarker;
  end;

  TTextDocumentAction = ( taDeleting, taDeleted, taChanging, taChanged, taInserting, taInserted );

  TTextDocument = class;

  // A notification event is fired before and after an operation that
  // deletes, changes or inserts a line.
  // Note that if a single operation affects several lines then
  // a separate event is fired for each line in unspecified order.
  TTextDocumentEvent = procedure ( Sender: TTextDocument; Action: TTextDocumentAction; Row: integer ) of object;

  TTextDocument = class
    FLines: TStringList;
    FOnChange: TTextDocumentEvent;

    procedure Notify( AAction: TTextDocumentAction; ARow: integer );
    procedure ChangeLine( ARow: integer; ALine: string );
    procedure DeleteLine( ARow: integer );
    procedure InsertLine( ARow: integer; ALine: string );
  public
    constructor Create;
    destructor Destroy; override;

    procedure AppendEmptyLine;

    // Actions: return true if an action is performed, false otherwise (e.g. because of invalid input)
    function DeleteChar( ACol, ARow: integer ): boolean;
    function DeleteBlock( APos1, APos2: TTextPosition ): boolean;
    function DeleteLines( ARow, ACount: integer ): boolean;
    function JoinLines( ARow: integer ): boolean;
    function SplitLine( ACol, ARow: integer ): boolean;
    function InsertChar( ACol, ARow: integer; AChar: char; AOverwrite: boolean ): boolean;
    function InsertText( ACol, ARow: integer; AText: string ): boolean;
    function InsertBlock( APos: TTextPosition; ABlock: TStringList ): boolean;
    function InsertLines( ARow: integer; ABlock: TStringList; ABlockRow, ABlockCount: integer ): boolean;

    // Utilities
    function CharAt( ACol, ARow: integer ): char;
    function IsBlockValid( const APos1, APos2: TTextPosition ): boolean;
    function IsPosValid( ACol, ARow: integer ): boolean;
    function IsEmpty: boolean;

    property Lines: TStringList read FLines;

    // Notification event
    property OnChange: TTextDocumentEvent read FOnChange write FOnChange;
  end;

  TTextEditor = class
    FDoc: TTextDocument;
    FCurCol: integer;
    FCurRow: integer;
    FOverwrite: boolean;

    function GetCurChar: char;
    function GetCurLine: string;
    function GetCurPos: TTextPosition;
    function GetLines: TStringList;
    procedure SetCurCol( AValue: integer );
    procedure SetCurRow( AValue: integer );
    procedure SetCurLine( const AValue: string );

  public
    constructor Create;
    destructor Destroy; override;

    // Cursor movement
    procedure Go( ACol, ARow: integer ); virtual;
    procedure GoFromCurrent( ADeltaCol, ADeltaRow: integer ); virtual;

    // Actions
    procedure Clear; virtual;
    procedure DoBackspace; virtual;
    procedure DoDelete; virtual;
    procedure DoReturn; virtual;
    procedure DoType( AChar: char ); virtual;

    property CurCol: integer read FCurCol write SetCurCol;
    property CurRow: integer read FCurRow write SetCurRow;
    property CurPos: TTextPosition read GetCurPos;
    property CurChar: char read GetCurChar;
    property CurLine: string read GetCurLine write SetCurLine;
    property Doc: TTextDocument read FDoc;
    property Lines: TStringList read GetLines;
    property Overwrite: boolean read FOverwrite write FOverwrite;
  end;

  TFullTextEditorAction = (
    eaUp, eaDown, eaLeft, eaRight,
    eaPageUp, eaPageDown,
    eaLineStart, eaLineEnd,
    eaPageTop, eaPageBottom,
    eaDocStart, eaDocEnd,
    eaScrollUp, eaScrollDown, eaScrollLeft, eaScrollRight,
    eaWordLeft, eaWordRight );

  TFullTextEditor = class(TTextEditor)
    FSelection: TTextBlock;
    FSelectionMarked: boolean;
    FPagePos: TTextPosition;
    FPageWidth: integer;
    FPageHeight: integer;
    FTabSize: integer;
    FVirtCol: integer;
    FLastSearchText: string;
    FLastSearchPos: TTextPosition;
    FSearchMatchCase: boolean;
    FSearchWholeWord: boolean;

    function GetLines: TStringList;
    function GetObjects( AIndex: integer ): TObject;
    procedure SetLines( AValue: TStringList );
    procedure SetObjects( AIndex: integer; AValue: TObject );
    procedure SetPageCol( AValue: integer );
    procedure SetPageRow( AValue: integer );
    procedure SetPagePos( AValue: TTextPosition );
    procedure BeginMoveCursor( AMarkSelection: boolean );
    procedure EndMoveCursor( AMarkSelection: boolean );

  protected
    function IsWordChar( AChar: char ): boolean; virtual;
    function SearchFrom( AText: string; var APos: TTextPosition ): boolean;
    procedure ForceCursorInPage;

  public
    constructor Create;
    destructor Destroy; override;

    function AbsColToVirtual( ACol, ARow: integer ): integer;
    function VirtualColToAbs( ACol, ARow: integer ): integer;

    procedure Clear; override;

    function SelectionLength: integer;
    function SelectionAsText: string;

    // Cursor
    procedure Go( ACol, ARow: integer ); override;
    procedure GoFromCurrent( ADeltaCol, ADeltaRow: integer ); override;
    procedure MoveTo( ACol, ARow: integer; AIsVirtual: boolean = false );

    // Actions
    function DoDeleteSelection: boolean; virtual;
    procedure DoInsertText( AText: string ); virtual;
    procedure DoBackspace; override;
    procedure DoDelete; override;
    procedure DoReturn; override;
    procedure DoTab( AExpand: boolean ); virtual;
    procedure DoSetCursor( ACol, ARow: integer; AMarkSelection: boolean );
    procedure DoAction( AAction: TFullTextEditorAction; AMarkSelection: boolean ); virtual;
    procedure DoType( AChar: char ); override;
    procedure DoSelectAll;
    procedure DoSelectWord( ACol, ARow: integer ); virtual;

    // Search
    function Search( AText: string ): boolean; virtual;
    function SearchAgain: boolean;

    property Lines: TStringList read GetLines write SetLines;
    property Objects[ AIndex: integer ]: TObject read GetObjects write SetObjects;
    property PageCol: integer read FPagePos.Col write SetPageCol;
    property PageRow: integer read FPagePos.Row write SetPageRow;
    property PagePos: TTextPosition read FPagePos write SetPagePos;
    property PageWidth: integer read FPageWidth write FPageWidth;
    property PageHeight: integer read FPageHeight write FPageHeight;
    property Selection: TTextBlock read FSelection;
    property SearchMatchCase: boolean read FSearchMatchCase write FSearchMatchCase;
    property SearchWholeWord: boolean read FSearchWholeWord write FSearchWholeWord;
    property SearchText: string read FLastSearchText;
    property TabSize: integer read FTabSize write FTabSize;
    property VirtCurCol: integer read FVirtCol;
  end;

function TextPosition( ACol, ARow: integer ): TTextPosition;
function ComparePositions( const APos1, APos2: TTextPosition ): integer;

implementation

uses
  SysUtils, Utils;

const
  ChBackspace = #$08;
  ChTab = #$09;
  ChNewLine = #$0A;
  ChReturn = #$0D;

function TextPosition( ACol, ARow: integer ): TTextPosition;
begin
  Result.Col := ACol;
  Result.Row := ARow;
end;

function ComparePositions( const APos1, APos2: TTextPosition ): integer;
begin
  Result := APos1.Row - APos2.Row;
  if( Result = 0 ) then
    Result := APos1.Col - APos2.Col;
end;

constructor TTextBlock.Create;
begin
  FMode := bmLinear;
  FData := nil;
  Clear;
end;

procedure TTextBlock.Clear;
begin
  StartMarker := TextPosition(0,0);
end;

function TTextBlock.Valid: boolean;
begin
  if( FMode = bmLinear ) then begin
    Result := (FStart.Row < FEnd.Row)or((FStart.Row = FEnd.Row)and(FStart.Col < FEnd.Col));
  end
  else begin
    Result := (FStart.Col < FEnd.Col)and(FStart.Row <= FEnd.Row);
  end;
end;

function TTextBlock.FirstRow: integer;
begin
  Result := -1;
  if( Valid ) then
    Result := FStart.Row;
end;

function TTextBlock.LastRow: integer;
begin
  Result := -1;
  if( Valid ) then
    Result := FEnd.Row;
end;

procedure TTextBlock.SetStartMarker( APos: TTextPosition );
begin
  FMarkerStart := APos;

  // All other block delimiters are same as the start marker at this time
  FMarkerEnd := FMarkerStart;
  FStart := FMarkerStart;
  FEnd := FMarkerStart;
end;

procedure TTextBlock.SetEndMarker( APos: TTextPosition );
begin
  FMarkerEnd := APos;

  if( ComparePositions( FMarkerStart, FMarkerEnd ) < 0 ) then begin
    FStart := FMarkerStart;
    FEnd := FMarkerEnd;
  end
  else begin
    FStart := FMarkerEnd;
    FEnd := FMarkerStart;
  end;
end;

constructor TTextDocument.Create;
begin
  FLines := TStringList.Create;
  FOnChange := nil;
  AppendEmptyLine;
end;

destructor TTextDocument.Destroy;
begin
  FLines.Free;
  inherited Destroy;
end;

procedure TTextDocument.Notify( AAction: TTextDocumentAction; ARow: integer );
begin
  if( Assigned(FOnChange) ) then
    FOnChange( Self, AAction, ARow );
end;

procedure TTextDocument.ChangeLine( ARow: integer; ALine: string );
begin
  Notify( taChanging, ARow );
  FLines[ ARow ] := ALine;
  Notify( taChanged, ARow );
end;

procedure TTextDocument.DeleteLine( ARow: integer );
begin
  Notify( taDeleting, ARow );
  FLines.Delete( ARow );
  Notify( taDeleted, ARow );
end;

procedure TTextDocument.InsertLine( ARow: integer; ALine: string );
begin
  Notify( taInserting, ARow );
  if( ARow >= FLines.Count ) then
    FLines.Add( ALine )
  else
    FLines.Insert( ARow, ALine );
  Notify( taInserted, ARow );
end;

function TTextDocument.CharAt( ACol, ARow: integer ): char;
begin
  Result := #0;
  if( IsPosValid( ACol, ARow ) ) then begin
    if( ACol < Length(FLines[ARow]) ) then
      Result := FLines[ARow][ACol+1]
  end;
end;

procedure TTextDocument.AppendEmptyLine;
begin
  InsertLine( FLines.Count, '' );
end;

function TTextDocument.IsEmpty: boolean;
begin
  Result := (FLines.Count = 1)and(Length(FLines[0]) = 0);
end;

function TTextDocument.DeleteChar;
var
  S: string;
begin
  Result := IsPosValid( ACol, ARow );
  if( Result ) then begin
    S := FLines[ ARow ];
    if( Length(S) <= ACol ) then begin
      if( ARow+1 < FLines.Count ) then begin
        // Join line to next
        FLines[ ARow ] := S + StringOfChar( ' ', ACol-Length(S) );
        JoinLines( ARow ); // Note: this will also notify the above change
      end;
    end
    else begin
      // Delete character at cursor
      Delete( S, ACol+1, 1 );
      ChangeLine( ARow, S );
    end
  end;
end;

function TTextDocument.JoinLines( ARow: integer ): boolean;
begin
  Result := (ARow >= 0 )and( ARow+1 < FLines.Count );
  if( Result ) then begin
    ChangeLine( ARow, FLines[ARow] + FLines[ARow+1] );
    DeleteLines( ARow+1, 1 );
  end;
end;

function TTextDocument.SplitLine( ACol, ARow: integer ): boolean;
begin
  Result := IsPosValid( ACol, ARow );
  if( Result ) then begin
    Inc( ACol ); // Make index string-friendly
    InsertLine( ARow+1, Copy(FLines[ARow], ACol, Length(FLines[ARow])) );
    ChangeLine( ARow, Copy( FLines[ARow], 1, ACol-1 ) );
  end;
end;

function TTextDocument.InsertChar;
var
  S: string;
begin
  Result := IsPosValid( ACol, ARow );
  if( Result ) then begin
    S := FLines[ ARow ];
    if( Length(S) <= ACol ) then begin
      // Character must be appended past end of line
      S := S + StringOfChar( ' ', ACol-Length(S) ) + AChar;
    end
    else if( AOverwrite ) then begin
      // Character is replaced
      S[ACol+1] := AChar;
    end
    else begin
      // Character is inserted
      Insert( AChar, S, ACol+1 );
    end;
    ChangeLine( ARow, S );
  end;
end;

function TTextDocument.InsertText( ACol, ARow: integer; AText: string ): boolean;
var
  S: string;
begin
  Result := IsPosValid( ACol, ARow );
  if( Result ) then begin
    S := FLines[ ARow ];
    if( Length(S) <= ACol ) then begin // Append past end of line
      S := S + StringOfChar( ' ', ACol-Length(S) ) + AText;
    end
    else begin
      Insert( AText, S, ACol+1 );
    end;
    ChangeLine( ARow, S );
  end;
end;

function TTextDocument.IsPosValid( ACol, ARow: integer ): boolean;
begin
  Result := (ACol >= 0)and( ARow >= 0 )and( ARow < FLines.Count );
end;

function TTextDocument.IsBlockValid( const APos1, APos2: TTextPosition ): boolean;
begin
  Result :=
    IsPosValid( APos1.Col, APos1.Row )and
    IsPosValid( APos2.Col, APos2.Row )and
    (ComparePositions( APos1, APos2 ) < 0);
end;

function TTextDocument.DeleteBlock( APos1, APos2: TTextPosition ): boolean;
var
  S: string;
begin
  Result := IsBlockValid( APos1, APos2 );
  if( Result ) then begin
    if( APos1.Row = APos2.Row ) then begin
      // Block is on a single line
      S := FLines[ APos1.Row ];
      Delete( S, APos1.Col+1, APos2.Col-APos1.Col );
      ChangeLine( APos1.Row, S );
    end
    else begin
      // First line: delete from start of block to end of line
      S := FLines[ APos1.Row ];
      Delete( S, APos1.Col+1, Length(S) );
      ChangeLine( APos1.Row, S );
      // Main block: delete all lines included between the first and the last
      DeleteLines( APos1.Row+1, APos2.Row-APos1.Row-1 );
      // Last line: delete from start of line to end of block
      APos2.Row := APos1.Row+1; // Take into account the previous deletions
      if( APos2.Col > 0 ) then begin
        S := FLines[ APos2.Row ];
        Delete( S, 1, APos2.Col );
        ChangeLine( APos2.Row, S );
      end;
      // Join first and last line
      JoinLines( APos1.Row );
    end;
  end;
end;

function TTextDocument.DeleteLines( ARow, ACount: integer ): boolean;
begin
  Result := false;
  if( ARow >= 0 ) then begin
    while( ACount > 0 )and( ARow < FLines.Count ) do begin
      DeleteLine( ARow );
      Dec( ACount );
      Result := true;
    end;
    if( FLines.Count = 0 ) then
      AppendEmptyLine;
  end;
end;

function TTextDocument.InsertLines( ARow: integer; ABlock: TStringList; ABlockRow, ABlockCount: integer ): boolean;
var
  I: integer;
begin
  Result := false;
  if( ARow >= 0 )and( ARow < FLines.Count ) then begin
    I := ABlockRow;
    while( I < ABlock.Count )and( ABlockCount > 0 ) do begin
      InsertLine( ARow, ABlock[I] );
      Inc( ARow );
      Inc( I );
      Dec( ABlockCount );
      Result := true;
    end;
  end;
end;

function TTextDocument.InsertBlock( APos: TTextPosition; ABlock: TStringList ): boolean;
begin
  Result := IsPosValid( APos.Col, APos.Row )and( ABlock.Count > 0 );
  if( Result ) then begin
    SplitLine( APos.Col, APos.Row );
    ChangeLine(APos.Row, FLines[APos.Row] + ABlock[0] );
    InsertLines( APos.Row+1, ABlock, 1, ABlock.Count-1 );
  end;
end;

procedure TTextEditor.Go( ACol, ARow: integer );
begin
  CurCol := ACol;
  CurRow := ARow;
end;

procedure TTextEditor.GoFromCurrent( ADeltaCol, ADeltaRow: integer );
begin
  CurCol := CurCol + ADeltaCol;
  CurRow := CurRow + ADeltaRow;
end;

constructor TTextEditor.Create;
begin
  FDoc := TTextDocument.Create;
  FCurCol := 0;
  FCurRow := 0;
end;

destructor TTextEditor.Destroy;
begin
  FDoc.Free;
  inherited Destroy;
end;

procedure TTextEditor.Clear;
begin
  FDoc.Lines.Clear;
  FCurCol := 0;
  FCurRow := 0;
end;

function TTextEditor.GetCurChar: char;
begin
  Result := FDoc.CharAt( FCurCol, FCurRow );
end;

function TTextEditor.GetCurLine: string;
begin
  Result := FDoc.Lines[ FCurRow ];
end;

function TTextEditor.GetCurPos: TTextPosition;
begin
  Result := TextPosition( FCurCol, FCurRow );
end;

function TTextEditor.GetLines: TStringList;
begin
  Result := FDoc.Lines;
end;

procedure TTextEditor.SetCurCol( AValue: integer );
begin
  if( AValue < 0 ) then AValue := 0;
  FCurCol := AValue;
end;

procedure TTextEditor.SetCurRow( AValue: integer );
begin
  if( AValue < 0 ) then AValue := 0;
  if( AValue >= FDoc.Lines.Count ) then AValue := FDoc.Lines.Count-1;
  FCurRow := AValue;
end;

procedure TTextEditor.SetCurLine( const AValue: string );
begin
  FDoc.Lines[ FCurRow ] := AValue;
end;

procedure TTextEditor.DoBackspace;
begin
  if( FCurCol > 0 ) then begin // Back one character
    GoFromCurrent( -1, 0 );
    if( FCurCol+1 <= Length(CurLine) ) then
      FDoc.DeleteChar( FCurCol, FCurRow );
  end
  else if( FCurRow > 0 ) then begin // Join the line above
    Go( Length(FDoc.Lines[FCurRow-1]), FCurRow-1 );
    FDoc.JoinLines( FCurRow );
  end;
end;

procedure TTextEditor.DoDelete;
begin
  // Delete char under cursor
  FDoc.DeleteChar( FCurCol, FCurRow );
end;

procedure TTextEditor.DoReturn;
begin
  if( Overwrite ) then begin
    if( FCurRow+1 >= FDoc.Lines.Count ) then
      FDoc.AppendEmptyLine;
    GoFromCurrent( 0, +1 );
  end
  else begin
    FDoc.SplitLine( FCurCol, FCurRow );
    Go( 0, FCurRow+1 );
  end;
end;

procedure TTextEditor.DoType( AChar: char );
begin
  FDoc.InsertChar( FCurCol, FCurRow, AChar, FOverwrite );
  GoFromCurrent( +1, 0 );
end;

constructor TFullTextEditor.Create;
begin
  inherited Create;
  FSelection := TTextBlock.Create;
  FSelectionMarked := false;
  FPagePos := TextPosition( 0, 0 );
  FPageWidth := 80;
  FPageHeight := 25;
  FTabSize := 0;
  FVirtCol := 0;
  FLastSearchText := '';
  FLastSearchPos := TextPosition( 0, 0 );
  FSearchMatchCase := false;
  FSearchWholeWord := false;
end;

destructor TFullTextEditor.Destroy;
begin
  FDoc.Free;
  FSelection.Free;
  inherited Destroy;
end;

procedure TFullTextEditor.Clear;
begin
  inherited Clear;
  FSelection.Clear;
  FPagePos := TextPosition( 0, 0 );
end;

function TFullTextEditor.GetLines: TStringList;
begin
  Result := FDoc.Lines;
end;

procedure TFullTextEditor.SetLines( AValue: TStringList );
begin
  FDoc.Lines.Assign( AValue );
  if( Lines.Count = 0 ) then
    FDoc.AppendEmptyLine;
  FSelection.Clear;
  FPagePos := TextPosition( 0, 0 );
  MoveTo( 0, 0 );
end;

function TFullTextEditor.GetObjects( AIndex: integer ): TObject;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < Lines.Count ) then
    Result := Lines.Objects[ AIndex ];
end;

procedure TFullTextEditor.SetObjects( AIndex: integer; AValue: TObject );
begin
  if( AIndex >= 0 )and( AIndex < Lines.Count ) then
    Lines.Objects[ AIndex ] := AValue;
end;

procedure TFullTextEditor.SetPageCol( AValue: integer );
begin
  SetPagePos( TextPosition( AValue, FPagePos.Row ) );
end;

procedure TFullTextEditor.SetPageRow( AValue: integer );
begin
  SetPagePos( TextPosition( FPagePos.Col, AValue ) );
end;

procedure TFullTextEditor.SetPagePos( AValue: TTextPosition );
var
  Col, Row: integer;
begin
  Col := AValue.Col;
  Row := AValue.Row;

  if( Col < 0 ) then Col := 0;
  if( Row < 0 ) then Row := 0;
  if( Row >= Lines.Count ) then Row := Lines.Count-1;

  if( Col <> FPagePos.Col )or( Row <> FPagePos.Row ) then begin
    FPagePos.Col := Col;
    FPagePos.Row := Row;
  end;
end;

function TFullTextEditor.IsWordChar( AChar: char ): boolean;
begin
  Result := AChar in ['a'..'z','A'..'Z','_'];
end;

procedure TFullTextEditor.ForceCursorInPage;
var
  PageCol, PageRow: integer;
  Col, Row: integer;
begin
  // Make sure cursor visible within page
  PageCol := FPagePos.Col;
  PageRow := FPagePos.Row;
  Col := FVirtCol;
  Row := CurRow;

  if( Row < PageRow ) then
    PageRow := Row;
  if( Row >= PageRow + FPageHeight ) then
    PageRow := Row - FPageHeight + 1;
  if( Col < PageCol ) then
    PageCol := Col;
  if( Col >= PageCol + FPageWidth ) then
    PageCol := Col - FPageWidth + 1;

  PagePos := TextPosition( PageCol, PageRow );
end;

procedure TFullTextEditor.MoveTo( ACol, ARow: integer; AIsVirtual: boolean );
begin
  // Clip cursor coordinates in the valid range
  if( ARow < 0 ) then ARow := 0;
  if( ARow >= FDoc.Lines.Count ) then ARow := FDoc.Lines.Count-1;
  if( ACol < 0 ) then ACol := 0;

  // Get absolute position if called in virtual coordinates
  if( AIsVirtual ) then
    ACol := VirtualColToAbs( ACol, ARow );

  // Set cursor position if different from current
  if( CurCol <> ACol )or( CurRow <> ARow ) then begin
    inherited Go( ACol, ARow );
  end;

  // Update virtual cursor position
  FVirtCol := AbsColToVirtual( ACol, ARow );

  ForceCursorInPage;
end;

procedure TFullTextEditor.Go( ACol, ARow: integer );
begin
  MoveTo( ACol, ARow );
end;

procedure TFullTextEditor.GoFromCurrent( ADeltaCol, ADeltaRow: integer );
begin
  MoveTo( CurCol+ADeltaCol, CurRow+ADeltaRow );
end;

procedure TFullTextEditor.BeginMoveCursor( AMarkSelection: boolean );
begin
  if( AMarkSelection )and( not FSelectionMarked ) then begin
    FSelection.StartMarker := TextPosition( CurCol, CurRow );
    FSelectionMarked := true;
  end;
end;

procedure TFullTextEditor.EndMoveCursor( AMarkSelection: boolean );
begin
  if( AMarkSelection ) then begin
    FSelection.EndMarker := TextPosition( CurCol, CurRow );
  end
  else begin
    FSelection.Clear;
    FSelectionMarked := false;
  end;
end;

procedure TFullTextEditor.DoAction( AAction: TFullTextEditorAction; AMarkSelection: boolean );
var
  Col, Row: integer;
procedure GoPrevWord;
  begin
    while( Col > 0 )and not IsWordChar(Doc.CharAt(Col-1,Row)) do
      Dec( Col );
    while( Col > 0 )and IsWordChar(Doc.CharAt(Col-1,Row)) do
      Dec( Col );
    MoveTo( Col, Row );
  end;
procedure GoNextWord;
  var
    MaxCol: integer;
  begin
    MaxCol := Length( CurLine );
    while( Col < MaxCol )and IsWordChar(Doc.CharAt(Col,Row)) do
      Inc( Col );
    while( Col < MaxCol )and not IsWordChar(Doc.CharAt(Col,Row)) do
      Inc( Col );
    MoveTo( Col, Row );
  end;
procedure GoRow( ARow: integer );
  begin
    MoveTo( FVirtCol, ARow, true );
  end;
begin
  // Save current cursor position
  Col := CurCol;
  Row := CurRow;

  BeginMoveCursor( AMarkSelection );

  case AAction of
    eaUp: GoRow( Row-1 );
    eaDown: GoRow( Row+1 );
    eaLeft: MoveTo( Col-1, Row );
    eaRight: MoveTo( Col+1, Row );
    eaPageUp: begin PageRow := PageRow - FPageHeight; GoRow( Row-FPageHeight ); end;
    eaPageDown: begin PageRow := PageRow + FPageHeight; GoRow( Row+FPageHeight ); end;
    eaLineStart: MoveTo( 0, Row );
    eaLineEnd: MoveTo( Length( FDoc.Lines[Row] ), Row );
    eaPageTop: GoRow( FPagePos.Row );
    eaPageBottom: GoRow( FPagePos.Row + FPageHeight-1 );
    eaDocStart: MoveTo( 0, 0 );
    eaDocEnd: MoveTo( Length( FDoc.Lines[FDoc.Lines.Count-1] ), FDoc.Lines.Count-1 );
    eaScrollUp: begin PageRow := PageRow - 1; if( CurRow >= PageRow+PageHeight ) then GoRow( PageRow+PageHeight-1 ); end;
    eaScrollDown: begin PageRow := PageRow + 1; if( CurRow < PageRow ) then GoRow( PageRow ); end;
    eaScrollLeft: begin PageCol := PageCol - 1; if( FVirtCol >= PageCol+PageWidth ) then FVirtCol := PageCol+PageWidth-1; end;
    eaScrollRight: begin PageCol := PageCol + 1; if( FVirtCol < PageCol ) then FVirtCol := PageCol; end;
    eaWordLeft: GoPrevWord;
    eaWordRight: GoNextWord;
  end;

  EndMoveCursor( AMarkSelection );
end;

procedure TFullTextEditor.DoSetCursor( ACol, ARow: integer; AMarkSelection: boolean );
begin
  BeginMoveCursor( AMarkSelection );
  Go( ACol, ARow );
  EndMoveCursor( AMarkSelection );
end;

function TFullTextEditor.DoDeleteSelection: boolean;
begin
  // Delete selection
  Result := FSelection.Valid;
  if( Result ) then begin
    Doc.DeleteBlock( FSelection.TopLeft, FSelection.BottomRight );
    Go( FSelection.TopLeft.Col, FSelection.TopLeft.Row );
    FSelection.Clear;
    FSelectionMarked := false;
  end;
end;

procedure TFullTextEditor.DoType( AChar: char );
var
  B: boolean;
begin
  B := DoDeleteSelection;

  case AChar of
    ChBackspace: if( not B ) then inherited DoBackspace;
    ChReturn: inherited DoReturn;
    else if( AChar = ChTab )or( AChar >= ' ' ) then inherited DoType( AChar );
  end;
end;

procedure TFullTextEditor.DoBackspace;
begin
  DoType( ChBackspace );
end;

procedure TFullTextEditor.DoDelete;
begin
  if( not DoDeleteSelection ) then
    inherited DoDelete;
end;

procedure TFullTextEditor.DoReturn;
begin
  DoType( ChReturn );
end;

procedure TFullTextEditor.DoTab( AExpand: boolean );
var
  T: integer;
begin
  // Get size of this tab
  T := FTabSize;
  if( T = 0 ) then T := 4;
  Dec( T, FVirtCol mod T );

  if( FOverwrite ) then begin
    BeginMoveCursor( false );
    MoveTo( FVirtCol + T, CurRow, true );
    EndMoveCursor( false );
  end
  else begin
    if( AExpand ) then begin
      DoDeleteSelection;
      DoInsertText( StringOfChar( ' ', T ) );
    end
    else begin
      DoType( ChTab );
    end;
  end;
end;

procedure TFullTextEditor.DoSelectWord( ACol, ARow: integer );
begin
  if( IsWordChar(Doc.CharAt(ACol,ARow)) ) then begin
    while( ACol < Length(Lines[ARow]) )and IsWordChar(Doc.CharAt(ACol,ARow)) do
      Inc( ACol );
    DoSetCursor( ACol, ARow, false );
    DoAction( eaWordLeft, true );
  end;
end;

procedure TFullTextEditor.DoSelectAll;
var
  N: integer;
begin
  N := FDoc.Lines.Count-1;
  FSelection.StartMarker := TextPosition( 0, 0 );
  FSelection.EndMarker := TextPosition( Length(FDoc.Lines[N]), N );
end;

const
  CRLF: string = #13#10;

procedure TFullTextEditor.DoInsertText( AText: string );
var
  Col, Row: integer;
  P, Q: integer;
begin
  Col := CurCol;
  Row := CurRow;
  repeat
    P := Pos( #10, AText );
    if( P = 0 ) then begin
      if( AText <> '' ) then
        Doc.InsertText( Col, Row, AText );
    end
    else begin
      // Insert partial string, then split
      Q := P-1;
      if( Q >= 1 )and( AText[Q] = #13 ) then
        Dec( Q );
      if( Q > 0 ) then begin
        Doc.InsertText( Col, Row, Copy(AText, 1, Q) );
        Inc( Col, Q );
      end;
      Doc.SplitLine( Col, Row );
      Col := 0;
      Inc( Row );
      // Delete the inserted string from the text
      Delete( AText, 1, P );
    end;
  until( P = 0 );

  Go( Col, Row );
end;

// Returns the length of the selected text in characters,
// assuming that each line is terminated by a CR-LF pair
function TFullTextEditor.SelectionLength: integer;
var
  I: integer;
begin
  Result := 0;
  if( Selection.Valid ) then begin
    // Now handle the first and last rows
    if( Selection.FirstRow = Selection.LastRow ) then begin
      Inc( Result, Selection.BottomRight.Col - Selection.TopLeft.Col );
    end
    else begin
      // First row
      Inc( Result, Length(CRLF) + Length(Lines[Selection.FirstRow]) - Selection.TopLeft.Col );
      // Main body
      for I:=Selection.FirstRow+1 to Selection.LastRow-1 do
        Inc( Result, Length(CRLF) + Length(Lines[I]) );
      // Last row
      Inc( Result, Selection.BottomRight.Col );
    end;
  end;
end;

function TFullTextEditor.SelectionAsText: string;
var
  N: integer;
  I: integer;
  P: PChar;
function StrAppend( P: PChar; const S: string ): PChar;
  begin
    StrCopy( P, PChar(S) );
    Result := P + Length( S );
  end;
begin
  Result := '';
  N := SelectionLength;
  if( N > 0 ) then begin
    if( Selection.FirstRow = Selection.LastRow ) then begin
      // Just one line...
      Result := Copy( Lines[Selection.FirstRow], Selection.TopLeft.Col+1, N );
    end
    else begin
      // This is a multiline selection, prepare to handle a long text!
      SetLength( Result, N+1 );
      P := PChar( Result );
      // First row
      if( Selection.TopLeft.Col < Length(Lines[Selection.FirstRow]) ) then
        P := StrAppend( P, Copy(Lines[Selection.FirstRow], Selection.TopLeft.Col+1, Length(Lines[Selection.FirstRow])) );
      P := StrAppend( P, CRLF );
      // Selection body
      for I:=Selection.FirstRow+1 to Selection.LastRow-1 do begin
        P := StrAppend( P, Lines[I] );
        P := StrAppend( P, CRLF );
      end;
      // Last row
      if( Selection.BottomRight.Col > 0 ) then
        StrAppend( P, Copy(Lines[Selection.LastRow], 0, Selection.BottomRight.Col) );
    end;
  end;
end;

function TFullTextEditor.AbsColToVirtual( ACol, ARow: integer ): integer;
var
  L: integer;
  S: string;
begin
  Result := ACol;

  if( FTabSize > 0 ) then begin
    Result := 0;
    S := FDoc.Lines[ARow];
    L := 0;
    while( L < ACol ) do begin
      if( L < Length(S) )and( S[L+1] = ChTab ) then
        Inc( Result, FTabSize - (Result mod FTabSize) )
      else
        Inc( Result );
      Inc( L );
    end;
  end;
end;

function TFullTextEditor.VirtualColToAbs( ACol, ARow: integer ): integer;
var
  I: integer;
  L: integer;
  S: string;
begin
  Result := ACol;

  if( FTabSize > 0 ) then begin
    Result := 0;
    S := FDoc.Lines[ARow];
    L := 0; // Virtual column
    while( L < ACol ) do begin
      // Get the size of this character
      I := 1;
      if( Result < Length(S) )and( S[Result+1] = ChTab ) then
        I := FTabSize - (L mod FTabSize);
      // Bump the absolute character count if virtual column is within the specified limit
      if( L+I <= ACol ) then
        Inc( Result );
      //
      Inc( L, I );
    end;
  end;
end;

function PosFrom( const Substr, Str: string; From: integer ): integer;
var
  P: PChar;
begin
  Result := 0;
  if( From-1+Length(Substr) <= Length(Str) ) then begin
    P := PChar(Str);
    Inc( P, From-1 );
    P := AnsiStrPos( P, PChar(Substr) );
    if( P <> nil ) then
      Result := P - PChar(Str) + 1;
  end;
end;

function TFullTextEditor.SearchFrom( AText: string; var APos: TTextPosition ): boolean;
var
  P: integer;
  S: string;
begin
  Result := false;

  if( AText = '' )or( not FDoc.IsPosValid( APos.Col, APos.Row ) ) then
    Exit;

  if( not FSearchMatchCase ) then
    AText := AnsiLowerCase(AText);

  while( APos.Row < FDoc.Lines.Count ) do begin
    if( FSearchMatchCase ) then
      S := FDoc.Lines[APos.Row]
    else
      S := AnsiLowerCase(FDoc.Lines[APos.Row]);
    P := PosFrom( AText, S, APos.Col+1 );
    if( P > 0 )and( FSearchWholeWord ) then begin
      if( (P > 1) and IsWordChar(S[P-1]) )or( (P+Length(AText) <= Length(S)) and IsWordChar(S[P+Length(AText)]) ) then begin
        // Text was found but it's not a whole word, keep searching
        APos.Col := P;
        continue;
      end;
    end;
    if( P > 0 ) then begin
      APos.Col := P-1;
      Result := true;
      break;
    end;
    APos.Col := 0;
    Inc( APos.Row );
  end;

  FLastSearchPos := APos;
  FLastSearchText := AText;
  if( Result ) then begin
    MoveTo( APos.Col, APos.Row );
    FSelection.StartMarker := APos;
    FSelection.EndMarker := TextPosition( APos.Col+Length(AText), APos.Row );
  end
  else begin
    FSelection.Clear;
  end;
end;

function TFullTextEditor.Search( AText: string ): boolean;
var
  P: TTextPosition;
begin
  P := TextPosition( 0, 0 );
  Result := SearchFrom( AText, P );
end;

function TFullTextEditor.SearchAgain: boolean;
var
  P: TTextPosition;
begin
  P := FLastSearchPos;
  Inc( P.Col );
  if( not FDoc.IsPosValid( P.Col, P.Row ) ) then
    P := TextPosition( 0, 0 );
  Result := SearchFrom( FLastSearchText, P );
end;

end.
