(*
    LCS - A study for a simple differencing program
    Full text editor window
    Copyright (c) 2002-2003,2004 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit FormEditor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, ExtCtrls, ToolWin,
  Editor, Syntax, UndoRedo, EditUndoBuffer, ImgList;

type
  TChangeLog = record
    FirstRow: integer;
    LastRow: integer;
    CursorPos: TTextPosition;
  end;

type
  TEditorForm = class(TForm)
    StatusBar: TStatusBar;
    PanelBottom: TPanel;
    HorzScrollBar: TScrollBar;
    VertScrollBar: TScrollBar;
    PaintBox: TPaintBox;
    ToolBar: TToolBar;
    TimerMouseMove: TTimer;
    ImageList: TImageList;
    BtnCut: TToolButton;
    BtnCopy: TToolButton;
    BtnPaste: TToolButton;
    PanelTopMargin: TPanel;
    ToolButton1: TToolButton;
    BtnUndo: TToolButton;
    BtnRedo: TToolButton;
    ToolButton2: TToolButton;
    BtnSearch: TToolButton;
    BtnSearchAgain: TToolButton;
    BtnOk: TToolButton;
    BtnCancel: TToolButton;
    ToolButton7: TToolButton;
    FindDialog: TFindDialog;
    PanelLeftBorder: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure PaintBoxPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure TimerMouseMoveTimer(Sender: TObject);
    procedure PaintBoxDblClick(Sender: TObject);
    procedure BtnCutClick(Sender: TObject);
    procedure HorzScrollBarChange(Sender: TObject);
    procedure VertScrollBarChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BtnOkClick(Sender: TObject);
    procedure BtnCancelClick(Sender: TObject);
    procedure FindDialogFind(Sender: TObject);
    procedure BtnSearchClick(Sender: TObject);
    procedure FindDialogClose(Sender: TObject);
    procedure BtnSearchAgainClick(Sender: TObject);
  private
    { Private declarations }
    FDoc: TFullTextEditor; // Document being edited
    FSyntaxDoc: TList; // Syntax hilight document
    FCaretHidden: boolean; // Caret hidden/visible flag
    FRecalcPageSize: boolean; // If true then page size is recomputed at next paint
    FCharWidth: integer; // Character font width (fixed font!)
    FCharHeight: integer; // Character font height (fixed font!)
    FSyntaxTokenizer: TSyntaxTokenizer; // Syntax tokenizer for highlight
    FChangeLog: TChangeLog; // Logs edit changes for undo
    FUndoBuffer: TUndoRedoBuffer; // Undo buffer
    FCurrentUndoItem: TUndoRedoItemList; // If non-nil logs edit events for undo
    FExpandTab: boolean; // Tab expansion flag
    FNextClipboardViewer: THandle; // To manage the clipboard viewers chain
    FCaptionHeader: string; // Original (design time) caption
    FPaintBitmap: TBitmap; // Bitmap for painting a text line offscreen

    function GetLines: TStringList;
    function GetTabSize: integer;
    procedure SetLines( AValue: TStringList );
    procedure SetTabSize( AValue: integer );

    function ResetChangeLogging: boolean;
    function FlushChangeLogging: boolean;

    procedure CopyToClipboard;
    procedure CutToClipboard;
    procedure DoHideCaret;
    procedure DoShowCaret;
    procedure MousePosToTextPos( var X, Y: integer );
    procedure OnChangeClipboardChain( var Msg: TMessage ); message WM_CHANGECBCHAIN;
    procedure OnChangeClipboardContent( var Msg: TMessage ); message WM_DRAWCLIPBOARD;
    procedure OnDocumentChanged( Sender: TTextDocument; Action: TTextDocumentAction; Row: integer );
    procedure OnEraseBackground( var Msg: TMessage ); message WM_ERASEBKGND;
    procedure OnCmDialogKey( var M: TCMDialogKey ); message CM_DIALOGKEY;
    procedure OnUndo( Sender: TObject; Item: TUndoRedoItem );
    procedure PaintLine( Canvas: TCanvas; Line: integer; ClipRect: TRect );
    procedure PasteFromClipboard;
    procedure StatusMessage( const Msg: string );
    procedure UpdateAfterSearch( Found: boolean );
    procedure UpdateControls;
    procedure UpdateDocumentAfterChanges;
    procedure UpdateScrollBarsSettings;
    procedure UpdateSyntaxDoc( AFromRow, AToRow: integer );
    procedure UpdateView;

  public
    { Public declarations }
    procedure Clear;
    procedure RebuildSyntaxDoc;
    procedure SetCaption( S: string );

    property Doc: TFullTextEditor read FDoc;
    property Lines: TStringList read GetLines write SetLines;
    property SyntaxTokenizer: TSyntaxTokenizer read FSyntaxTokenizer write FSyntaxTokenizer;
    property ExpandTab: boolean read FExpandTab write FExpandTab;
    property TabSize: integer read GetTabSize write SetTabSize;
  end;

var
  EditorForm: TEditorForm;

implementation

uses
  ClipBrd, Utils;

{$R *.DFM}

const
  InvalidValue = -1;

procedure TEditorForm.OnEraseBackground( var Msg: TMessage );
begin
  Msg.Result := 1;
end;

procedure TEditorForm.FormCreate(Sender: TObject);
begin
  FCaptionHeader := Caption;
  FDoc := TFullTextEditor.Create;
  FDoc.TabSize := 4;
  FDoc.Doc.OnChange := OnDocumentChanged;
  FSyntaxDoc := TList.Create;
  FCaretHidden := true;
  FRecalcPageSize := true;
  FCharWidth := InvalidValue;
  FCharHeight := InvalidValue;
  FSyntaxTokenizer := nil;
  FUndoBuffer := TUndoRedoBuffer.Create( 8 );
  FUndoBuffer.OnUndoItem := OnUndo;
  FUndoBuffer.OnRedoItem := OnUndo;
  FCurrentUndoItem := nil;
  FNextClipboardViewer := SetClipboardViewer( Handle );
  FExpandTab := false;
  FPaintBitmap := TBitmap.Create;
  FPaintBitmap.Canvas.Font.Assign( Font );
  UpdateControls;
  RebuildSyntaxDoc;
end;

procedure TEditorForm.FormDestroy(Sender: TObject);
begin
  ChangeClipboardChain( Handle, FNextClipboardViewer );
end;

procedure TEditorForm.SetCaption( S: string );
begin
  if( S = '' ) then
    S := '(no name)';
  Caption := FCaptionHeader + ' - ' + S;
end;

function TEditorForm.GetLines: TStringList;
begin
  Result := FDoc.Lines;
end;

procedure TEditorForm.SetLines( AValue: TStringList );
begin
  Clear;
  FDoc.Lines := AValue;
  ResetChangeLogging;
  UpdateControls;
  HorzScrollBar.Position := 0;
  VertScrollBar.Position := 0;
  UpdateScrollBarsSettings;
  RebuildSyntaxDoc;
end;

procedure TEditorForm.DoHideCaret;
begin
  if( not FCaretHidden ) then begin
    FCaretHidden := true;
    Windows.HideCaret(Handle);
  end;
end;

procedure TEditorForm.DoShowCaret;
var
  X, Y: integer;
begin
  X := FDoc.VirtCurCol - FDoc.PageCol;
  Y := FDoc.CurRow - FDoc.PageRow;
  if( Y >= 0 )and( Y < FDoc.PageHeight )and( X >= 0 )and( X < FDoc.PageWidth ) then begin
    SetCaretPos( PaintBox.Left + X*FCharWidth, PaintBox.Top + Y*FCharHeight );
    if( FCaretHidden ) then begin
      FCaretHidden := false;
      ShowCaret( Handle );
    end;
  end;
  StatusBar.Panels[0].Text := Format( ' Line:%d   Col:%d', [FDoc.CurRow+1,FDoc.VirtCurCol+1] );
end;

procedure TEditorForm.CopyToClipboard;
var
  S: string;
begin
  if( FDoc.SelectionLength > 0 ) then begin
    S := FDoc.SelectionAsText;
    Clipboard.SetTextBuf( PChar(S) );
  end;
end;

procedure TEditorForm.PasteFromClipboard;
begin
  if( Clipboard.HasFormat(CF_TEXT) ) then
    FDoc.DoDeleteSelection;
    FDoc.DoInsertText( Clipboard.AsText );
end;

procedure TEditorForm.CutToClipboard;
begin
  CopyToClipboard;
  FDoc.DoDeleteSelection;
end;

procedure TEditorForm.FormActivate(Sender: TObject);
begin
  CreateCaret( Handle, 0, 2, PaintBox.Canvas.TextHeight('0') );
  FCaretHidden := true;
end;

procedure TEditorForm.FormDeactivate(Sender: TObject);
begin
  DestroyCaret;
end;

procedure TEditorForm.UpdateDocumentAfterChanges;
begin
  HorzScrollBar.Position := FDoc.PageCol;
  VertScrollBar.Position := FDoc.PageRow;
  UpdateScrollBarsSettings;

  if( FlushChangeLogging ) then begin
    UpdateSyntaxDoc( FChangeLog.FirstRow, FChangeLog.LastRow );
  end;

  StatusMessage( '' );
  UpdateControls;
  UpdateView;
end;

procedure TEditorForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  ResetChangeLogging;

  if( Key in [#8,#13] )or( Key >= ' ' ) then begin
    FDoc.DoType( Key );
    UpdateDocumentAfterChanges;
  end
  else if( Key = #9 ) then begin
    FDoc.DoTab( FExpandTab );
    UpdateDocumentAfterChanges;
  end;
end;

procedure TEditorForm.OnCmDialogKey( var M: TCMDialogKey );
begin
  // This is enough to generate a key press event for VK_TAB!
  M.Result := 0;
end;

procedure TEditorForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  IsSelecting: boolean;
begin
  if( ssAlt in Shift ) then
    Exit;

  ResetChangeLogging;

  IsSelecting := ssShift in Shift; // Selection flag

  if( ssCtrl in Shift ) then case Key of
    Ord('C'),
    VK_INSERT:  CopyToClipboard;
    Ord('V'):   PasteFromClipboard;
    Ord('X'):   CutToClipboard;
    Ord('Y'):   if( FUndoBuffer.RedoCount > 0 ) then FUndoBuffer.Redo;
    Ord('Z'):   if( FUndoBuffer.UndoCount > 0 ) then FUndoBuffer.Undo;
    Ord('F'):   begin BtnSearchClick(Self); Exit; end;
    Ord('A'):   FDoc.DoSelectAll;
    VK_LEFT:    FDoc.DoAction( eaWordLeft, IsSelecting );
    VK_RIGHT:   FDoc.DoAction( eaWordRight, IsSelecting );
    VK_UP:      FDoc.DoAction( eaScrollUp, IsSelecting );
    VK_DOWN:    FDoc.DoAction( eaScrollDown, IsSelecting );
    VK_PRIOR:   FDoc.DoAction( eaPageTop, IsSelecting );
    VK_NEXT:    FDoc.DoAction( eaPageBottom, IsSelecting );
    VK_HOME:    FDoc.DoAction( eaDocStart, IsSelecting );
    VK_END:     FDoc.DoAction( eaDocEnd, IsSelecting );
    else
      Exit;
  end
  else case Key of
    VK_ESCAPE:  ModalResult := mrCancel;
    VK_F3:      begin BtnSearchAgainClick(Self); Exit; end;
    VK_INSERT:  if( ssShift in Shift ) then PasteFromClipboard else FDoc.Overwrite := not FDoc.Overwrite;
    VK_DELETE:  if( ssShift in Shift ) then CutToClipboard else FDoc.DoDelete;
    VK_PRIOR:   FDoc.DoAction( eaPageUp, IsSelecting );
    VK_NEXT:    FDoc.DoAction( eaPageDown, IsSelecting );
    VK_LEFT:    FDoc.DoAction( eaLeft, IsSelecting );
    VK_RIGHT:   FDoc.DoAction( eaRight, IsSelecting );
    VK_UP:      FDoc.DoAction( eaUp, IsSelecting );
    VK_DOWN:    FDoc.DoAction( eaDown, IsSelecting );
    VK_HOME:    FDoc.DoAction( eaLineStart, IsSelecting );
    VK_END:     FDoc.DoAction( eaLineEnd, IsSelecting );
    else
      Exit;
  end;

  UpdateDocumentAfterChanges;
end;

procedure TEditorForm.PaintLine( Canvas: TCanvas; Line: integer; ClipRect: TRect );
var
  I: integer;
  X: integer;
  L: TSyntaxLine;
  T: TSyntaxTokenItem;
begin
  X := -FDoc.PageCol*FCharWidth;
  L := TSyntaxLine( FDoc.Objects[Line] );
  if( L = nil ) then begin
    Canvas.TextRect( ClipRect, X, ClipRect.Top, ExpandTabs(FDoc.Lines[Line],FDoc.TabSize) );
  end
  else begin
    Canvas.FillRect( ClipRect );
    for I:=1 to L.TokenCount do begin
      T := L.Tokens[I-1];
      case T.Kind of
        tkComment: begin
          Canvas.Font.Style := [fsItalic];
          if( Canvas.Font.Color <> clHighlightText ) then Canvas.Font.Color := clNavy;
        end;
        tkString: begin
          Canvas.Font.Style := [];
          if( Canvas.Font.Color <> clHighlightText ) then Canvas.Font.Color := clTeal;
        end;
        tkIdentifier: begin
          Canvas.Font.Style := [fsBold];
          if( Canvas.Font.Color <> clHighlightText ) then Canvas.Font.Color := clBlack;
        end;
        else begin
          Canvas.Font.Style := [];
          if( Canvas.Font.Color <> clHighlightText ) then Canvas.Font.Color := clBlack;
        end;
      end;
      Canvas.TextRect( ClipRect, X, ClipRect.Top, Copy(L.Line, T.Pos, T.Len) );
      Inc( X, T.Len*FCharWidth );

      if( X > ClipRect.Right ) then
        break;

      if( X > ClipRect.Left ) then
        ClipRect.Left := X;
    end;
  end;
end;

procedure TEditorForm.PaintBoxPaint(Sender: TObject);
var
  I, L: integer;
  R: TRect;
begin
  with PaintBox do begin
    if( FCharWidth = InvalidValue )or( FCharHeight = InvalidValue ) then begin
      FCharWidth := Canvas.TextWidth('0');
      FCharHeight := Canvas.TextHeight('0');
    end;
    if( FRecalcPageSize ) then begin
      FRecalcPageSize := false;
      FDoc.PageWidth := Width div FCharWidth;
      FDoc.PageHeight := Height div FCharHeight;
      UpdateScrollBarsSettings;
    end;

    DoHideCaret;

    // Make sure bitmap has the right size
    if( FPaintBitmap.Width <> Width ) then
      FPaintBitmap.Width := Width;
    if( FPaintBitmap.Height <> FCharHeight ) then
      FPaintBitmap.Height := FCharHeight;

    for I:=0 to FDoc.PageHeight do begin // Draw one extra line!
      L := FDoc.PageRow + I;
      R := Rect( 0, 0, Width, FCharHeight );
      if( L < FDoc.Lines.Count ) then begin
        if( L > FDoc.Selection.FirstRow )and( L < FDoc.Selection.LastRow ) then begin
          // Fully selected text
          FPaintBitmap.Canvas.Brush.Color := clHighlight;
          FPaintBitmap.Canvas.Font.Color := clHighlightText;
          PaintLine( FPaintBitmap.Canvas, L, R );
        end
        else begin
          // Unselected or partially selected text
          FPaintBitmap.Canvas.Brush.Color := clWindow;
          FPaintBitmap.Canvas.Font.Color := clBlack;
          PaintLine( FPaintBitmap.Canvas, L, R );
          // If partially selected text, redraw the selection
          if( FDoc.Selection.Valid )and( L = FDoc.Selection.FirstRow )or( L = FDoc.Selection.LastRow ) then begin
            FPaintBitmap.Canvas.Brush.Color := clHighlight;
            FPaintBitmap.Canvas.Font.Color := clHighlightText;
            if( L = FDoc.Selection.FirstRow ) then
              R.Left := FDoc.AbsColToVirtual(FDoc.Selection.TopLeft.Col,FDoc.Selection.FirstRow)*FCharWidth;
            if( L = FDoc.Selection.LastRow ) then
              R.Right := FDoc.AbsColToVirtual(FDoc.Selection.BottomRight.Col,FDoc.Selection.LastRow)*FCharWidth;
            OffsetRect( R, -FDoc.PageCol*FCharWidth, 0 );
            PaintLine( FPaintBitmap.Canvas, L, R );
          end;
        end;
        Canvas.Draw( 0, I*FCharHeight, FPaintBitmap );
      end
      else begin
        // Background
        Canvas.Brush.Color := clWindow;
        Canvas.FillRect( Rect(0,I*FCharHeight,Width,(I+1)*FCharHeight) );
      end;
    end;

    DoShowCaret;
  end;
end;

function TEditorForm.ResetChangeLogging: boolean;
begin
  Result := true;
  FChangeLog.FirstRow := InvalidValue;
  FChangeLog.LastRow := InvalidValue;
  FChangeLog.CursorPos := TextPosition( FDoc.CurCol, FDoc.CurRow );
  FCurrentUndoItem.Free;
  FCurrentUndoItem := nil;
end;

function TEditorForm.FlushChangeLogging: boolean;
begin
  Result := true;

  if( FCurrentUndoItem <> nil ) then begin
    // Add cursor/block info
    FCurrentUndoItem.Add( TEditInfoUndoItem.Create( FChangeLog.CursorPos ) );

    if( FCurrentUndoItem.Tag = 0 ) then begin
      FUndoBuffer.Push( FCurrentUndoItem );
    end;
    // ...else it is a redo item and we do not need to do anything

    FCurrentUndoItem := nil;
  end;
end;

procedure TEditorForm.OnDocumentChanged( Sender: TTextDocument; Action: TTextDocumentAction; Row: integer );
begin
  if( FCurrentUndoItem = nil ) then begin
    // Create undo item if needed
    FCurrentUndoItem := TUndoRedoItemList.Create;
  end;

  if( Action = taDeleting ) then begin
    TSyntaxLine(Sender.Lines.Objects[Row]).Free;
    FCurrentUndoItem.Add( TEditUndoItem.Create( Row, Action, Sender.Lines[Row] ) );
  end
  else if( Action = taDeleted ) then begin
  end
  else if( Action = taChanging ) then begin
    FCurrentUndoItem.Add( TEditUndoItem.Create( Row, Action, Sender.Lines[Row] ) );
  end
  else if( Action = taChanged ) then begin
  end
  else if( Action = taInserted ) then begin
    FCurrentUndoItem.Add( TEditUndoItem.Create( Row, Action, '' ) );
  end;

  // Update line pointers
  if( FChangeLog.FirstRow = InvalidValue )or( Row < FChangeLog.FirstRow ) then
    FChangeLog.FirstRow := Row;

  if( FChangeLog.LastRow = InvalidValue )or( Row > FChangeLog.LastRow ) then
    FChangeLog.LastRow := Row;
end;

procedure TEditorForm.UpdateSyntaxDoc( AFromRow, AToRow: integer );
var
  Row1, Row2: integer;
  Flags: TSyntaxLineFlags;
  SL: TSyntaxLine;
begin
  if( FSyntaxTokenizer = nil )or( AFromRow = InvalidValue )or( AToRow = InvalidValue ) then
    Exit;

  Row1 := AFromRow;
  if( Row1 < 0 ) then Row1 := 0;
  Row2 := AToRow;
  if( Row2 < 0 )or( Row2 >= FDoc.Lines.Count-1 ) then Row2 := FDoc.Lines.Count-1;

  Flags := [];
  if( Row1 > 0 ) then
    Flags := TSyntaxLine(FDoc.Objects[Row1-1]).Flags;

  // Update the changed block
  while( Row1 <= Row2 ) do begin
    SL := TSyntaxLine( FDoc.Objects[Row1] );
    if( SL = nil ) then begin
      SL := TSyntaxLine.Create( FSyntaxTokenizer );
      FDoc.Objects[Row1] := SL;
    end;
    SL.Assign( ExpandTabs(FDoc.Lines[Row1],FDoc.TabSize), Flags );
    Flags := SL.Flags;
    Inc( Row1 );
  end;

  // Now we need to keep updating lines until the current status
  // matches that of the following line
  while( Row1 < FDoc.Lines.Count ) do begin
    SL := TSyntaxLine( FDoc.Objects[Row1] );
    if( Flags = SL.Flags ) then
      break;
    SL.Assign( FDoc.Lines[Row1], Flags );
    Flags := SL.Flags;
    Inc( Row1 );
  end;
end;

procedure TEditorForm.RebuildSyntaxDoc;
begin
  UpdateSyntaxDoc( 0, FDoc.Lines.Count-1 );
end;

procedure TEditorForm.Clear;
var
  I: integer;
begin
  for I:=1 to FDoc.Lines.Count do
    TSyntaxLine(FDoc.Lines.Objects[I-1]).Free;
  FDoc.Clear;
end;

procedure TEditorForm.FormResize(Sender: TObject);
begin
  FRecalcPageSize := true;
end;

const
  LeftButtonDown: boolean = false;
  SkipButtonDown: boolean = false;

procedure TEditorForm.MousePosToTextPos( var X, Y: integer );
begin
  X := X div FCharWidth + FDoc.PageCol;
  Y := Y div FCharHeight + FDoc.PageRow;
  if( Y >= FDoc.Lines.Count ) then Y := FDoc.Lines.Count-1;
  X := FDoc.VirtualColToAbs( X, Y );
end;

procedure TEditorForm.PaintBoxMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if( SkipButtonDown ) then begin
    SkipButtonDown := false;
    Exit;
  end;

  if( Button = mbLeft ) then begin
    LeftButtonDown := true;
    MousePosToTextPos( X, Y );
    FDoc.DoSetCursor( X, Y, ssShift in Shift );
    UpdateControls;
    UpdateView;
  end;
end;

procedure TEditorForm.PaintBoxMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if( LeftButtonDown ) then begin
    MousePosToTextPos( X, Y );
    if( X <> FDoc.CurCol )or( Y <> FDoc.CurRow ) then begin
      FDoc.DoSetCursor( X, Y, true );
      UpdateControls;
      UpdateView;
    end;
  end;
end;

procedure TEditorForm.PaintBoxMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if( Button = mbLeft ) then begin
    LeftButtonDown := false;
  end;
end;

procedure TEditorForm.TimerMouseMoveTimer(Sender: TObject);
var
  P: TPoint;
begin
  if( LeftButtonDown ) then begin
    GetCursorPos( P );
    P := PaintBox.ScreenToClient( P );
    PaintBoxMouseMove( Sender, [], P.X, P.Y );
  end;
end;

procedure TEditorForm.PaintBoxDblClick(Sender: TObject);
var
  P: TPoint;
begin
  SkipButtonDown := true; // Do not let the MouseDown event interfere with the double click
  GetCursorPos( P );
  P := PaintBox.ScreenToClient( P );
  MousePosToTextPos( P.X, P.Y );
  FDoc.DoSelectWord( P.X, P.Y );
  UpdateControls;
  UpdateView;
end;

procedure TEditorForm.OnUndo( Sender: TObject; Item: TUndoRedoItem );
var
  L: TUndoRedoItemList;
  E: TEditUndoItem;
  I: integer;
begin
  // Copy the undo items into another list in reverse order
  L := TUndoRedoItemList.Create;
  for I:=TUndoRedoItemList(Item).Count downto 1 do begin
    L.Add( TUndoRedoItemList(Item).Item[I-1] );
  end;

  // Now the undo/redo manager is going to automatically move this item
  // into the redo buffer, so we use the item itself to record the
  // change we make during the undo, obtaining the redo for free!
  Assert( FCurrentUndoItem = nil );
  ResetChangeLogging;
  FCurrentUndoItem := Item as TUndoRedoItemList;
  FCurrentUndoItem.Clear;
  FCurrentUndoItem.Tag := 1; // Do not manually push this item!

  // Perform the undo actions (reminder: we are logging these actions
  // into the undo item itself for the redo, that's why we use a copy
  // of the item list)
  for I:=1 to L.Count do begin
    if( L.Item[I-1] is TEditUndoItem ) then begin
      E := TEditUndoItem( L.Item[I-1] );
      case E.Action of
        taDeleting: begin
          FDoc.Doc.InsertLine( E.Row, E.Line );
        end;
        taChanging: begin
          FDoc.Doc.ChangeLine( E.Row, E.Line );
        end;
        taInserted: begin
          FDoc.Doc.DeleteLine( E.Row );
        end;
      end
    end
    else with L.Item[I-1] as TEditInfoUndoItem do begin
      FDoc.Go( CurCol, CurRow );
    end;
  end;

  L.Free;
end;

procedure TEditorForm.BtnCutClick(Sender: TObject);
begin
  ResetChangeLogging;
  with Sender as TToolButton do begin
    case Tag of
      1: CutToClipboard;
      2: CopyToClipboard;
      3: PasteFromClipboard;
      4: if( FUndoBuffer.UndoCount > 0 ) then FUndoBuffer.Undo;
      5: if( FUndoBuffer.RedoCount > 0 ) then FUndoBuffer.Redo;
    end;
  end;
  UpdateDocumentAfterChanges;
end;

procedure TEditorForm.UpdateControls;
const
  OverwriteText: array[boolean] of string = ('Insert','Overwrite');
begin
  BtnUndo.Enabled := FUndoBuffer.UndoCount > 0;
  BtnRedo.Enabled := FUndoBuffer.RedoCount > 0;
  BtnCopy.Enabled := FDoc.Selection.Valid;
  BtnCut.Enabled := FDoc.Selection.Valid;
  StatusBar.Panels[1].Text := OverwriteText[ FDoc.Overwrite ];
end;

procedure TEditorForm.UpdateView;
begin
  PaintBoxPaint(Self);
end;

procedure TEditorForm.HorzScrollBarChange(Sender: TObject);
begin
  if( FDoc.PageCol <> HorzScrollBar.Position ) then begin
    FDoc.PageCol := HorzScrollBar.Position;
    UpdateView;
  end;
end;

procedure TEditorForm.VertScrollBarChange(Sender: TObject);
begin
  if( FDoc.PageRow <> VertScrollBar.Position ) then begin
    FDoc.PageRow := VertScrollBar.Position;
    UpdateView;
  end;
end;

procedure TEditorForm.UpdateScrollBarsSettings;
var
  M: integer;
begin
  M := FDoc.Lines.Count-1;
  if( VertScrollBar.Max <> M ) then
    VertScrollBar.Max := M;
  if( VertScrollBar.LargeChange <> FDoc.PageHeight ) then
    VertScrollBar.LargeChange := FDoc.PageHeight;
  if( VertScrollBar.PageSize <> FDoc.PageHeight ) then
    VertScrollBar.PageSize := FDoc.PageHeight;
  if( HorzScrollBar.LargeChange <> FDoc.PageWidth ) then
    HorzScrollBar.LargeChange := FDoc.PageWidth;
  if( HorzScrollBar.PageSize <> FDoc.PageWidth ) then
    HorzScrollBar.PageSize := FDoc.PageWidth;
end;

procedure TEditorForm.OnChangeClipboardChain( var Msg: TMessage );
begin
  if( FNextClipboardViewer = THandle(Msg.WParam) ) then
    FNextClipboardViewer := Msg.LParam
  else if( FNextClipboardViewer <> 0 ) then
    SendMessage( FNextClipboardViewer, Msg.Msg, Msg.WParam, Msg.LParam );
  Msg.Result := 0;
end;

procedure TEditorForm.OnChangeClipboardContent( var Msg: TMessage );
begin
  BtnPaste.Enabled := Clipboard.HasFormat(CF_TEXT);
  if( FNextClipboardViewer <> 0 ) then
    SendMessage( FNextClipboardViewer, Msg.Msg, Msg.WParam, Msg.LParam );
  Msg.Result := 0;
end;

procedure TEditorForm.BtnOkClick(Sender: TObject);
begin
  ModalResult := mrOk;
end;

procedure TEditorForm.BtnCancelClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TEditorForm.FindDialogFind(Sender: TObject);
begin
  FindDialog.CloseDialog;
  FDoc.SearchMatchCase := frMatchCase in FindDialog.Options;
  FDoc.SearchWholeWord := frWholeWord in FindDialog.Options;
  UpdateAfterSearch( FDoc.Search( FindDialog.FindText ) );
end;

procedure TEditorForm.BtnSearchClick(Sender: TObject);
begin
  FormDeactivate(Self);
  FindDialog.Execute;
end;

procedure TEditorForm.FindDialogClose(Sender: TObject);
begin
  FormActivate(Self);
  DoShowCaret;
end;

procedure TEditorForm.BtnSearchAgainClick(Sender: TObject);
begin
  UpdateAfterSearch( FDoc.SearchAgain );
end;

procedure TEditorForm.StatusMessage( const Msg: string );
begin
  StatusBar.Panels[2].Text := Msg;
end;

procedure TEditorForm.UpdateAfterSearch(Found: boolean);
begin
  if( not Found ) then
    StatusMessage( Format( ' Search text ''%s'' not found', [FDoc.SearchText] ) )
  else
    StatusMessage( '' );
  UpdateControls;
  UpdateView;
end;

function TEditorForm.GetTabSize: integer;
begin
  Result := FDoc.TabSize;
end;

procedure TEditorForm.SetTabSize( AValue: integer );
begin
  FDoc.TabSize := AValue;
end;

end.

