(*
    Generic undo/redo support.
    Copyright (c) 2003 Alessandro Scotti
    www.ascotti.org

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit UndoRedo;

interface

uses
  Classes, SysUtils;

const
  UnlimitedUndoRedo = -1;

type
  ERedoBufferEmpty = class(Exception);
  EUndoBufferEmpty = class(Exception);
  EUndoBufferFull = class(Exception);

type
  (* Base class for undo/redo items. *)
  TUndoRedoItem = class
    FTag: longint;
  public
    constructor Create;
    destructor Destroy; override;

    property Tag: longint read FTag write FTag;
  end;

type
  (* List of undo/redo items, for batch operations. *)
  TUndoRedoItemList = class(TUndoRedoItem)
    FList: TList;

    function GetCount: integer;
    function GetItem( AIndex: integer ): TUndoRedoItem;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Add( AItem: TUndoRedoItem );

    procedure Clear;

    property Item[AIndex: integer]: TUndoRedoItem read GetItem;
    property Count: integer read GetCount;
  end;

type
  TUndoRedoEvent = procedure( Sender: TObject; Item: TUndoRedoItem ) of object;

type
  (* Undo/redo buffer. *)
  TUndoRedoBuffer = class
    FItems: TList;
    FItemIndex: integer;
    FMaxItemCount: integer;
    FMultipleEventForItemList: boolean;
    FOnFree: TUndoRedoEvent;
    FOnRedo: TUndoRedoEvent;
    FOnUndo: TUndoRedoEvent;

    function GetRedoCount: integer;
    function GetRedoItems( AIndex: integer ): TUndoRedoItem;
    function GetUndoCount: integer;
    function GetUndoItems( AIndex: integer ): TUndoRedoItem;

    procedure FreeItem( AIndex: integer; ARemoveFromList: boolean );
    procedure NotifyEvent( AEventHandler: TUndoRedoEvent; AItem: TUndoRedoItem );

  public
    constructor Create( AMaxBufferLength: integer = UnlimitedUndoRedo );

    destructor Destroy; override;

    (*
        Removes all items from the buffer.

        Note: the OnFreeItem event is called for each undo/redo item
        before it is deleted.
    *)
    procedure Clear;

    (*
        Removes all items from the redo buffer.

        Note: the OnFreeItem event is called for each undo/redo item
        before it is deleted.
    *)
    procedure ClearRedo;

    (*
        Pushes an undo/redo item in the buffer.

        If the buffer is full (as set on construction) then an
        exception is raised.

        Note: the buffer takes full ownership of the undo/redo items
        it contains. This means that it will automatically free an
        item when it's no longer needed. Before deleting an item,
        however, the OnFreeItem event is called so that you can
        take opportune action such as for example release data associated
        with the item.
    *)
    procedure Push( AItem: TUndoRedoItem );

    (*
        Executes a redo operation.

        If the redo buffer is empty an exception is raised, otherwise
        the buffer is updated accordingly and the OnRedo event is
        called to execute whatever action is associated to the operation.
    *)
    procedure Redo;

    (*
        Executes an undo operation.

        If the undo buffer is empty an exception is raised, otherwise
        the buffer is updated accordingly and the OnUndo event is
        called to execute whatever action is associated to the operation.
    *)
    procedure Undo;

    (* Items in the redo buffer *)
    property RedoItems[AIndex: integer]: TUndoRedoItem read GetRedoItems;

    (* Items in the undo buffer *)
    property UndoItems[AIndex: integer]: TUndoRedoItem read GetUndoItems;

  published
    (*
        Called just before an undo/redo item is deleted.

        Note: as long as the undo/redo buffer is concerned,
        a TUndoRedoItemList item is just a single undo/redo item so
        this method is called just once for the entire list, and not
        for every item contained in the list itself.
    *)
    property OnFreeItem: TUndoRedoEvent read FOnFree write FOnFree;

    (* Called to redo an undo/redo item. *)
    property OnRedoItem: TUndoRedoEvent read FOnRedo write FOnRedo;

    (* Called to undo an undo/redo item. *)
    property OnUndoItem: TUndoRedoEvent read FOnUndo write FOnUndo;

    (*
        If true, undo or redo for an item list generates a different
        event for each item in the list (the event for the list
        itself is NOT generated in this case).
    *)
    property MultipleEventForItemList: boolean read FMultipleEventForItemList write FMultipleEventForItemList;

    (* Returns the number of items in the redo buffer *)
    property RedoCount: integer read GetRedoCount;

    (* Returns the number of items in the undo buffer *)
    property UndoCount: integer read GetUndoCount;
  end;

implementation

constructor TUndoRedoItem.Create;
begin
  inherited Create;
  FTag := 0;
end;

destructor TUndoRedoItem.Destroy;
begin
  inherited Destroy;
end;

constructor TUndoRedoItemList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TUndoRedoItemList.Destroy;
var
  I: integer;
begin
  for I:=1 to FList.Count do begin
    TUndoRedoItem( FList[I-1] ).Free;
  end;
  FList.Free;
  inherited Destroy;
end;

procedure TUndoRedoItemList.Add( AItem: TUndoRedoItem );
begin
  FList.Add( AItem );
end;

procedure TUndoRedoItemList.Clear;
begin
  FList.Clear;
end;

function TUndoRedoItemList.GetCount: integer;
begin
  Result := FList.Count;
end;

function TUndoRedoItemList.GetItem( AIndex: integer ): TUndoRedoItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < FList.Count ) then begin
    Result := TUndoRedoItem( FList[AIndex] );
  end;
end;

constructor TUndoRedoBuffer.Create( AMaxBufferLength: integer );
begin
  inherited Create;
  FItems := TList.Create;
  FItemIndex := 0;
  FMaxItemCount := AMaxBufferLength;
  FMultipleEventForItemList := false;
  FOnFree := nil;
  FOnRedo := nil;
  FOnUndo := nil;
end;

destructor TUndoRedoBuffer.Destroy;
begin
  Clear;
  FItems.Free;
  inherited Destroy;
end;

function TUndoRedoBuffer.GetRedoCount: integer;
begin
  Result := FItems.Count - FItemIndex;
end;

function TUndoRedoBuffer.GetRedoItems( AIndex: integer ): TUndoRedoItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < RedoCount ) then begin
    Result := TUndoRedoItem( FItems[FItemIndex+AIndex] );
  end;
end;

function TUndoRedoBuffer.GetUndoCount: integer;
begin
  Result := FItemIndex;
end;

function TUndoRedoBuffer.GetUndoItems( AIndex: integer ): TUndoRedoItem;
begin
  Result := nil;
  if( AIndex >= 0 )and( AIndex < UndoCount ) then begin
    Result := TUndoRedoItem( FItems[AIndex] );
  end;
end;

procedure TUndoRedoBuffer.FreeItem( AIndex: integer; ARemoveFromList: boolean );
var
  E: TUndoRedoItem;
begin
  if( AIndex >= 0 )and( AIndex < FItems.Count ) then begin
    E := TUndoRedoItem( FItems[AIndex] );

    if( Assigned( FOnFree ) ) then begin
      FOnFree( Self, E );
    end;

    E.Free;

    if( ARemoveFromList ) then begin
      FItems.Delete( AIndex );
    end;
  end;
end;

procedure TUndoRedoBuffer.Clear;
var
  I: integer;
begin
  for I:=1 to FItems.Count do begin
    FreeItem( I-1, false );
  end;

  FItems.Clear;
  FItemIndex := 0;
end;

procedure TUndoRedoBuffer.ClearRedo;
begin
  // Remove all items past the current index
  while( FItems.Count > FItemIndex ) do begin
    FreeItem( FItemIndex, true );
  end;
end;

procedure TUndoRedoBuffer.Push( AItem: TUndoRedoItem );
begin
  // Remove all items in the redo buffer
  ClearRedo;

  // Check item count
  while( FMaxItemCount > 0 )and( FItemIndex >= FMaxItemCount ) do begin
    FreeItem( 0, true );
    Dec( FItemIndex );
  end;

  // Add the undo/redo item
  FItems.Add( AItem );
  Inc( FItemIndex );
end;

procedure TUndoRedoBuffer.NotifyEvent( AEventHandler: TUndoRedoEvent; AItem: TUndoRedoItem );
var
  I: integer;
  L: TUndoRedoItemList;
begin
  if( Assigned(AEventHandler) ) then begin
    if( FMultipleEventForItemList )and( AItem is TUndoRedoItemList ) then begin
      L := TUndoRedoItemList( AItem );
      for I:=1 to L.Count do begin
        AEventHandler( Self, L.Item[I-1] );
      end;
    end
    else begin
      AEventHandler( Self, AItem );
    end;
  end;
end;

procedure TUndoRedoBuffer.Redo;
begin
  if( RedoCount = 0 ) then begin
    raise ERedoBufferEmpty.Create( 'The redo buffer is empty.' );
  end;

  Inc( FItemIndex );

  NotifyEvent( FOnRedo, TUndoRedoItem(FItems[FItemIndex-1]) );
end;

procedure TUndoRedoBuffer.Undo;
begin
  if( UndoCount = 0 ) then begin
    raise EUndoBufferEmpty.Create( 'The undo buffer is empty.' );
  end;

  Dec( FItemIndex );

  NotifyEvent( FOnUndo, TUndoRedoItem(FItems[FItemIndex]) );
end;

end.
